use std::sync::Once;

use axum::{
    body::Body,
    extract::State,
    http::{header, Response, StatusCode, Uri},
    response::IntoResponse,
    routing::get,
};
use axum_controller::*;
use hypertext::{maud, GlobalAttributes, Renderable};
use rust_embed::RustEmbed;

mod html_elements {
    use hypertext::elements;
    pub use hypertext::html_elements::*;

    elements! {
        bla {
            blub
        }
        my_element {
            my_attribute
        }
    }
}

struct UiController {}

#[controller(
    state = AppState
)]
impl UiController {
    #[route(GET "/")]
    async fn index(State(_): State<AppState>) -> impl IntoResponse {
        maud! {
            html lang="en" {
                head {
                    meta charset="UTF-8";
                    meta name="viewport" content="width=device-width, initial-scale=1.0";
                    title {
                        "LLM Chat App"
                    }
                    script type="module" src="/dist/datastar.min.js" {}
                    link rel="stylesheet" href="/dist/styles.min.css";
                    link rel="icon" href="/dist/favicon.ico";
                }
                body class="bg-gray-100" {
                    div class="container mx-auto p-4" {
                        h1 class="text-2xl font-bold mb-4" {
                            "LLM Chat App"
                        }
                        div class="bg-white p-6 rounded-lg shadow-md" {
                            div id="chat" class="mb-4" {
                                // Chat messages will appear here
                            }
                            form id="chat-form" {
                                textarea id="user-input" class="w-full p-2 border rounded-lg mb-2" placeholder="Type your message..." {}
                                button type="submit" class="bg-blue-500 text-white p-2 rounded-lg" {
                                    "Send"
                                }
                            }
                        }
                    }
                    script {
                        (hypertext::Raw("
                            console.log(\"asd\");
                            document.getElementById('chat-form').addEventListener('submit', function(event) {
                                event.preventDefault();
                                const userInput = document.getElementById('user-input').value;
                                const chatContainer = document.getElementById('chat');
                                chatContainer.innerHTML += `<div class='mb-2'><strong>You:</strong> ${userInput}</div>`;
                                document.getElementById('user-input').value = '';
                                console.log(\"asd\");

                                // Mock response from LLM
                                setTimeout(() => {
                                    chatContainer.innerHTML += `<div class='mb-2'><strong>LLM:</strong> This is a mock response.</div>`;
                                }, 1000);
                            });
                        "))
                    }
                }
            }
        }
        .render()
    }
}

struct DistController;

#[controller(state=AppState, path="/dist")]
impl DistController {
    #[route(GET "/*path")]
    async fn static_handler(path: String, _: State<AppState>) -> impl IntoResponse {
        let path = path.trim_start_matches('/').to_string();

        StaticFile(path)
    }
}

fn markup_404(uri: String) -> impl Renderable {
    maud! {
        h1 { "404" }
        p { (uri) " Not Found" }
        @for i in 0..5 {
            div .{"m-" (i)} { (i) }
        }
    }
}

fn markup_405() -> impl Renderable {
    maud! {
        h1 { "405" }
        p { "Method not allowed!" }
        @for i in 1..3 {
            div .{"m-" (i)} {}
        }
    }
}

// Finally, we use a fallback route for anything that didn't match.
async fn handle_404(uri: Uri) -> impl IntoResponse {
    (
        StatusCode::NOT_FOUND,
        axum::response::Html(markup_404(format!("{uri:?}")).render()),
    )
        .into_response()
}

async fn handle_405() -> impl IntoResponse {
    (
        StatusCode::METHOD_NOT_ALLOWED,
        axum::response::Html(markup_405().render()),
    )
        .into_response()
}

pub struct StaticFile<T>(pub T);

impl<T> IntoResponse for StaticFile<T>
where
    T: Into<String>,
{
    fn into_response(self) -> Response<Body> {
        let path = self.0.into();
        tracing::debug!(?path);
        #[derive(RustEmbed)]
        #[folder = "public/"]
        struct Asset;

        match Asset::get(path.as_str()) {
            Some(content) => {
                let mime = mime_guess::from_path(path).first_or_octet_stream();

                ([(header::CONTENT_TYPE, mime.as_ref())], content.data).into_response()
            }
            None => (
                StatusCode::NOT_FOUND,
                axum::response::Html(markup_404(path).render()),
            )
                .into_response(),
        }
    }
}

pub fn initialize_logger() {
    static INIT: Once = Once::new();

    INIT.call_once(|| {
        let env_filter = tracing_subscriber::EnvFilter::builder()
            .with_default_directive(tracing::level_filters::LevelFilter::INFO.into())
            .from_env_lossy();

        tracing_subscriber::fmt()
            .pretty()
            .with_env_filter(env_filter)
            .init();
    });
}

#[derive(Clone, Debug)]
struct AppState {
    _field: String,
}

#[tokio::main]
async fn main() {
    initialize_logger();

    let app_state = AppState {
        _field: "".to_string(),
    };

    // TODO pick free port/config
    let listener = tokio::net::TcpListener::bind("0.0.0.0:8000").await.unwrap();

    let router: axum::Router = axum::Router::new()
        .merge(UiController::into_router(app_state.clone()))
        .merge(DistController::into_router(app_state.clone()))
        .fallback_service(get(handle_404))
        .method_not_allowed_fallback(handle_405)
        .with_state(app_state);

    let _ = axum::serve(listener, router.into_make_service()).await;
}
