use leptos::*;
use leptos_router::*;
use serde::{Deserialize, Serialize};

use crate::app::components::{svgs::*, FormControl};

#[derive(PartialEq, Debug, Clone, Serialize, Deserialize, sqlx::FromRow)]
pub struct BackendSettings {
    version: i64, // auto-increments, the current settings are the highest version
    model_dir: String,
    default_model: String,
    created_at: chrono::NaiveDateTime,
}

/// Returns the initial default directory for model files. ($PWD/models)
#[server]
async fn default_model_dir() -> Result<String, ServerFnError> {
    let pwd = std::env::current_dir()?.to_string_lossy().to_string();
    let default_dir = format!("{}/models", pwd);
    Ok(default_dir)
}

/// Fetches the latest backend settings from the database.
#[server]
async fn load_backend_settings() -> Result<BackendSettings, ServerFnError> {
    use crate::server::pool;

    tracing::debug!(msg = "Load settings");

    let pool = pool()?;

    let settings = sqlx::query_as!(
        BackendSettings,
        "SELECT version, model_dir, default_model, created_at FROM settings ORDER BY version DESC \
         LIMIT 1"
    )
    .fetch_optional(&pool)
    .await?;

    match settings {
        Some(settings) => Ok(settings),
        None => Err(ServerFnError::ServerError("Settings not found".to_string())),
    }
}

/// Updates the backend settings with new model directory and default model.
#[server(UpdateSettings)]
async fn update_settings(model_dir: String, default_model: String) -> Result<(), ServerFnError> {
    use crate::server::pool;

    tracing::debug!(msg = "Updating settings", ?model_dir, ?default_model);

    let pool = pool()?;

    let result = {
        sqlx_macros::expand_query!(
            source = "INSERT INTO settings (model_dir, default_model) VALUES (?, ?)",
            args = [model_dir, default_model,]
        )
    }
    .execute(&pool)
    .await?;

    if result.rows_affected() == 0 {
        return Err(ServerFnError::ServerError(
            "Failed to create new settings".to_string(),
        ));
    }

    Ok(())
}

/// A form component for updating the backend settings.<|im_end|>
#[component]
pub fn SettingsForm() -> impl IntoView {
    let update_settings_action = create_server_action::<UpdateSettings>();
    let backend_settings_res = create_local_resource(
        || (),
        |_| async move {
            let settings = load_backend_settings().await;
            match settings {
                Ok(settings) => settings,
                Err(_) => {
                    let default_dir = default_model_dir().await.unwrap_or("".to_string());
                    BackendSettings {
                        version: 1,
                        model_dir: default_dir,
                        default_model: "meta-llama-3-8b-instruct-imat-IQ3_XXS.gguf".to_string(),
                        created_at: chrono::Utc::now().naive_utc(),
                    }
                }
            }
        },
    );

    let model_dir = move || {
        if let Some(setting) = backend_settings_res.get() {
            setting.model_dir.to_string()
        } else {
            "loading".to_string()
        }
    };

    let default_model = move || {
        if let Some(setting) = backend_settings_res.get() {
            setting.default_model.to_string()
        } else {
            "loading".to_string()
        }
    };

    view! {
        <ActionForm action=update_settings_action>
            <FormControl label="Default Model">
                <input
                    type="text"
                    name="default_model"
                    placeholder="meta-llama-3-8b-instruct-imat-IQ3_XXS.gguf"
                    value=default_model
                    class="grow max-w-[75%] input input-xs input-bordered"
                />
            </FormControl>
            <FormControl label="model_dir">
                <input
                    type="text"
                    name="model_dir"
                    placeholder="/opt/ai/models"
                    value=model_dir
                    class="grow max-w-[75%] input input-xs input-bordered"
                />
            </FormControl>
            <div class="w-full form-control">
                <button class="btn btn-primary" type="submit">
                    <SaveSvg />
                </button>
            </div>
        </ActionForm>
    }
}
