use async_trait::async_trait;
use futures::StreamExt;
use reqwest::Client;
use reqwest_eventsource::{Event, EventSource};
use serde::{Deserialize, Serialize};
use tokio::sync::mpsc;
use uuid::Uuid;

use super::ChatService;
use crate::api::{ChannelMessage, Chat, ChatMessage};

#[derive(Serialize, Debug)]
struct LlamaChatCompletionRequest {
    stream:   bool,
    model:    String,
    messages: Vec<LlamaChatMessage>,
}

impl From<Chat> for LlamaChatCompletionRequest {
    fn from(value: Chat) -> Self {
        Self {
            stream:   true,
            model:    "default".to_string(),
            messages: value.history.into_iter().map(|e| e.into()).collect(),
        }
    }
}

#[derive(Serialize, Debug)]
struct LlamaChatMessage {
    role:    String,
    content: String,
}

impl From<ChatMessage> for LlamaChatMessage {
    fn from(chat_message: ChatMessage) -> Self {
        Self {
            role:    chat_message.role.into(),
            content: chat_message.content,
        }
    }
}

#[derive(Deserialize, Debug)]
struct LlamaChatDelta {
    content: Option<String>,
}

#[derive(Deserialize, Debug)]
struct LlamaChatChoice {
    finish_reason: Option<String>,
    index: i64,
    delta: Option<LlamaChatDelta>,
}

#[derive(Deserialize, Debug)]
struct LlamaChatResponse {
    id: String,
    created: i64, // TODO Make DateTime (from unix)
    model: String,
    choices: Vec<LlamaChatChoice>,
    object: String,
}

pub struct LlamaService {
    id: Uuid,
}

impl LlamaService {
    pub fn new(id: Uuid) -> Self {
        Self {
            id,
        }
    }
}

#[tracing::instrument(level = "debug")]
async fn do_chat_request(chat: Chat, sender: mpsc::Sender<ChannelMessage>) -> anyhow::Result<()> {
    let client = Client::new();
    let request_body: LlamaChatCompletionRequest = chat.into();

    let request_builder = client
        // # .post("http://localhost:8080/v1/chat/completions")
        .post("http://100.64.0.3:18080/v1/chat/completions")
        .header("Content-Type", "application/json")
        .json(&request_body);

    let mut es = EventSource::new(request_builder)?;

    while let Some(event) = es.next().await {
        match event {
            Ok(Event::Open) => tracing::debug!("Connection Open!"),
            Ok(Event::Message(event)) => match event.event.as_str() {
                "message" => {
                    let data = event.data;
                    tracing::debug!(?data);
                    if data == "[DONE]" {
                        sender
                            .send(ChannelMessage::Stop)
                            .await
                            .expect("channel fail");

                        es.close();
                        break;
                    }

                    let response: LlamaChatResponse = serde_json::from_str(&data).expect("no json");

                    for choice in response.choices.into_iter() {
                        if let Some(delta) = choice.delta {
                            let content = delta.content;
                            if let Some(content_str) = content {
                                let mut string = content_str.clone();
                                tracing::debug!("STRING: {}", string);
                                if string == "\n" {
                                    string = "<br>".to_string()
                                }
                                let msg = ChannelMessage::Token(string);
                                tracing::debug!(?msg);
                                sender.send(msg.clone()).await.expect("sender fail");
                            }
                        }
                    }
                }
                _ => {
                    todo! {}
                }
            },
            Err(err) => {
                tracing::debug!("Error: {}", err);
                sender
                    .send(ChannelMessage::Stop)
                    .await
                    .expect("channel fail");
                // sender.send("STOP".to_string()).await.expect("sender fail");
                es.close();
            }
        }
    }

    Ok(())
}

#[async_trait]
impl ChatService for LlamaService {
    #[tracing::instrument(level = "debug", skip(self))]
    async fn chat_request(self, history: Vec<ChatMessage>, rx: mpsc::Sender<ChannelMessage>) {
        let chat = Chat::new(self.id, history, "".to_string());
        do_chat_request(chat, rx).await.expect("chat reqest error");
    }
}
